# import typing

# from alang.funcs import Function
# from alang.nodes import AccessMode, AddressSpace, CodeOptions, Module, Variable
# from alang.typs import Array, Struct, Vector, Tensor, int_type, float_type, tensor_type
# from alang.langs import get_language
# from alang.compiler import Compiler

if __name__ == "__main__":
    pass
